/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

#define FE_SMO_DEBUG	FALSE
#define FE_SMO_DUMP		FALSE

#include "operator/operator.pmh"

//* TODO check both inputs for normals
//* TODO dot check should be optional

namespace fe
{

void SurfaceMetricOp::initialize(void)
{
	catalog<String>("icon")="FE_beta";

	catalog<bool>("PartitionInput")=false;
	catalog<String>("PartitionInput","label")="Partition Input";
	catalog<bool>("PartitionInput","joined")=true;
	catalog<String>("PartitionInput","hint")=
			"Separate primitives into non-overlapping UV spaces.";

	catalog<String>("PartitionAttr")="mesh";
	catalog<String>("PartitionAttr","label")="Partition Attribute";
	catalog<String>("PartitionAttr","enabler")="PartitionInput";
	catalog<String>("PartitionAttr","hint")=
			"Primitive string attribute on primary input mesh"
			" describing membership in a non-overlapping UV space."
			"  This attribute will be transferred to the output points.";

	catalog<String>("ValueAttr")="value";
	catalog<String>("ValueAttr","label")="Value Attribute";
	catalog<String>("ValueAttr","hint")=
			"Name of point attribute to place results.";

	catalog<Real>("Minimum")= -1.0;
	catalog<Real>("Minimum","low")= -10.0;
	catalog<Real>("Minimum","high")=0.0;

	catalog<Real>("Maximum")=1.0;
	catalog<Real>("Maximum","high")=10.0;

	catalog<I32>("Width")=256;
	catalog<I32>("Width","high")=1024;

	catalog<I32>("Height")=256;
	catalog<I32>("Height","high")=1024;

	catalog< sp<Component> >("Input Surface");

	catalog< sp<Component> >("Contrasting Surface");

	//* don't copy input to output
	catalog< sp<Component> >("Output Surface");
	catalog<String>("Output Surface","IO")="output";
}

void SurfaceMetricOp::handle(Record& a_rSignal)
{
	const BWORD partitionInput=catalog<bool>("PartitionInput");
	const String partitionAttr=
			partitionInput? catalog<String>("PartitionAttr"): "";
	const String valueAttr=catalog<String>("ValueAttr");
	const Real minimum=catalog<Real>("Minimum");
	const Real maximum=catalog<Real>("Maximum");
	const I32 width=catalog<I32>("Width");
	const I32 height=catalog<I32>("Height");

	sp<DrawI> spDrawI;
	if(!accessDraw(spDrawI,a_rSignal)) return;

	sp<SurfaceAccessibleI> spOutputAccessible;
	if(!accessOutput(spOutputAccessible,a_rSignal)) return;

	sp<SurfaceAccessorI> spOutputUVs;
	if(!access(spOutputUVs,spOutputAccessible,e_point,"uv")) return;

	sp<SurfaceAccessorI> spOutputWidth;
	if(!access(spOutputWidth,spOutputAccessible,
			e_detail,"raster_width")) return;

	sp<SurfaceAccessorI> spOutputHeight;
	if(!access(spOutputHeight,spOutputAccessible,
			e_detail,"raster_height")) return;

	sp<SurfaceAccessorI> spOutputMin;
	if(!access(spOutputMin,spOutputAccessible,
			e_detail,"raster_min")) return;

	sp<SurfaceAccessorI> spOutputMax;
	if(!access(spOutputMax,spOutputAccessible,
			e_detail,"raster_max")) return;

	sp<SurfaceAccessorI> spOutputPartAttr;
	if(!access(spOutputPartAttr,spOutputAccessible,
			e_detail,"raster_partition")) return;

	sp<SurfaceAccessorI> spOutputValue;
	if(!access(spOutputValue,spOutputAccessible,e_point,valueAttr)) return;

	sp<SurfaceAccessibleI> spInputAccessible;
	if(!access(spInputAccessible,"Input Surface")) return;

	sp<SurfaceI> spContrast;
	if(!access(spContrast,"Contrasting Surface")) return;

	sp<SurfaceI> spInput;
	if(!access(spInput,spInputAccessible)) return;

	spInput->setRefinement(0);

	sp<SurfaceAccessorI> spOutputPartition;

	U32 mapCount=1;
	if(partitionInput)
	{
		spInput->partitionWith(partitionAttr);
		spInput->prepareForUVSearch();
		mapCount=spInput->partitionCount();

		if(!access(spOutputPartition,spOutputAccessible,
				e_point,partitionAttr)) return;
	}

#if FE_SMO_DUMP
		feLog("mapCount %d partition %d \"%s\"\n",
				mapCount,partitionInput,partitionAttr.c_str());
#endif

	spOutputWidth->set(0,width);
	spOutputHeight->set(0,height);
	spOutputMin->set(0,minimum);
	spOutputMax->set(0,maximum);
	spOutputPartAttr->set(0,partitionAttr);

	const Real maxDistance=maximum-minimum;

	U32 pointCount=0;

	//* TODO create each map in parallel

	for(U32 mapIndex=0;mapIndex<mapCount;mapIndex++)
	{
		String partition;
		if(partitionInput)
		{
			partition=spInput->partition(mapIndex);
			spInput->setPartitionFilter(partition);
		}

#if FE_SMO_DUMP
		feLog("map %d/%d partition \"%s\"\n",
				mapIndex,mapCount,partition.c_str());
#endif

		for(U32 vIndex=0;vIndex<height;vIndex++)
		{
			const Real vCoord=vIndex/(height-1.0);

			for(U32 uIndex=0;uIndex<width;uIndex++)
			{
				const Real uCoord=uIndex/(width-1.0);

				const Vector2 uv(uCoord,vCoord);
				const SpatialTransform sample=spInput->sample(uv);
				const SpatialVector point=sample.translation();
				const SpatialVector normal=sample.column(1);

				const SpatialVector origin=point+minimum*normal;
				Real value=maximum;

#if	FE_SMO_DEBUG
				feLog("\n%d,%d %s\n",uIndex,vIndex,partition.c_str());
#endif

				const U32 maxHits=2;
				std::vector< sp<SurfaceI::ImpactI> > impactArray=
						spContrast->rayImpacts(origin,normal,
						maxDistance,maxHits);
				const U32 hitCount=impactArray.size();
				for(U32 hitIndex=0;hitIndex<hitCount;hitIndex++)
				{
					sp<SurfaceI::ImpactI>& rspImpact=impactArray[hitIndex];

					const SpatialVector hitNormal=rspImpact->normal();

#if	FE_SMO_DEBUG
					feLog("metric hit %d/%d normal %s vs %s\n",
							hitIndex,hitCount,
							c_print(hitNormal),c_print(normal));
#endif

					const Real minDot=0.0;	//* TODO param
					if(dot(hitNormal,normal)>minDot)
					{
						value=rspImpact->distance()+minimum;

#if	FE_SMO_DEBUG
						feLog("  dist %.6G min %.6G value %.6G\n",
								rspImpact->distance(),minimum,value);
#endif
						break;
					}
				}

				spDrawI->drawPoints(&point,NULL,1,false,NULL);

				spOutputUVs->set(pointCount,uv);
				spOutputValue->set(pointCount,value);

				if(spOutputPartition.isValid())
				{
					spOutputPartition->set(pointCount,partition);
				}

				pointCount++;

#if FE_SMO_DUMP
				feLog("%3d ",I32(value*100));
#endif
			}
#if FE_SMO_DUMP
			feLog("\n");
#endif
		}
	}
}

} // namespace
