/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "surface.pmh"

namespace fe
{

sp<RecordArray> SurfaceAccessorBase::recordArray(void) const
{
	RecordGroup::iterator it=m_spRG->begin();
	if(it!=m_spRG->end())
	{
		return *it;
	}
	return sp<RecordArray>(NULL);
}

U32 SurfaceAccessorBase::count(void) const
{
	sp<RecordArray> spRecordArray=recordArray();

	return spRecordArray.isValid()? spRecordArray->length(): 0;
}

U32 SurfaceAccessorBase::subCount(U32 a_index) const
{
	if(m_attribute==SurfaceAccessibleI::e_vertices)
	{
		sp<RecordArray> spRecordArray=recordArray();
		if(spRecordArray.isValid())
		{
			Record record=spRecordArray->getRecord(a_index);
			std::vector<I32> vertexArray=m_vertexIndexAccessor(record);
			return vertexArray.size();
		}
	}

	return 0;
}

void SurfaceAccessorBase::ensureType(String a_attrType)
{
//	feLog("SurfaceAccessorBase::ensureType \"%s\" vs \"%s\" for \"%s\"\n",
//			a_attrType.c_str(),m_attrType.c_str(),m_attrName.c_str());

	if(a_attrType==m_attrType)
	{
		return;
	}
	if(m_attrType.empty())
	{
		sp<RecordArray> spRecordArray=recordArray();
		if(spRecordArray.isNull())
		{
			feLog("SurfaceAccessorBase::ensureType \"%s\" no records\n",
					a_attrType.c_str());
			return;
		}

		m_attrType=a_attrType;

		sp<Layout> spLayout=spRecordArray->layout();
		const String layoutName=spLayout->name();

//		feLog("SurfaceAccessorBase::ensureType \"%s\" for \"%s\" in \"%s\"\n",
//				a_attrType.c_str(),m_attrName.c_str(),layoutName.c_str());

		sp<Scope> spScope=spLayout->scope();

		spScope->declare(layoutName);
		spScope->support(m_attrName,m_attrType);
		spScope->populate(layoutName,m_attrName);

		if(m_attrType=="string")
		{
			m_accessorString.setup(spScope,m_attrName,m_attrType);
		}
		else if(m_attrType=="integer")
		{
			m_accessorInteger.setup(spScope,m_attrName,m_attrType);
		}
		else if(m_attrType=="real")
		{
			m_accessorReal.setup(spScope,m_attrName,m_attrType);
		}
		else if(m_attrType=="vector3")
		{
			m_accessorSpatialVector.setup(spScope,m_attrName,m_attrType);
		}

		return;
	}
	feLog("SurfaceAccessorBase::ensureType(\"%s\") already \"%s\"\n",
			a_attrType.c_str(),m_attrType.c_str());
}

void SurfaceAccessorBase::set(U32 a_index,U32 a_subIndex,String a_string)
{
	ensureType("string");

	sp<RecordArray> spRecordArray=recordArray();
	if(spRecordArray.isNull() || a_index>=spRecordArray->length())
	{
		return;
	}

	Record record=spRecordArray->getRecord(a_index);
	if(record.isValid())
	{
		m_accessorString(record)=a_string;
	}
}

String SurfaceAccessorBase::string(U32 a_index,U32 a_subIndex)
{
	ensureType("string");

	sp<RecordArray> spRecordArray=recordArray();
	if(spRecordArray.isNull() || a_index>=spRecordArray->length())
	{
		return "";
	}

	Record record=spRecordArray->getRecord(a_index);
	if(record.isValid())
	{
		return m_accessorString(record);
	}

	return "";
}

void SurfaceAccessorBase::set(U32 a_index,U32 a_subIndex,I32 a_integer)
{
	if(m_attribute==SurfaceAccessibleI::e_vertices)
	{
		FEASSERT(m_element==SurfaceAccessibleI::e_primitive);

		ensureType("intarray");

		sp<RecordArray> spRecordArray=recordArray();
		if(spRecordArray.isNull() || a_index>=spRecordArray->length())
		{
			return;
		}

		Record record=spRecordArray->getRecord(a_index);
		std::vector<I32> vertexArray=m_vertexIndexAccessor(record);

		if(a_subIndex>=vertexArray.size())
		{
			return;
		}

		vertexArray[a_subIndex]=a_integer;
	}

	ensureType("integer");

	sp<RecordArray> spRecordArray=recordArray();
	if(spRecordArray.isNull() || a_index>=spRecordArray->length())
	{
		return;
	}

	Record record=spRecordArray->getRecord(a_index);
	if(record.isValid())
	{
		m_accessorInteger(record)=a_integer;
	}
}

I32 SurfaceAccessorBase::integer(U32 a_index,U32 a_subIndex)
{
	if(m_attribute==SurfaceAccessibleI::e_vertices)
	{
		FEASSERT(m_element==SurfaceAccessibleI::e_primitive);

		ensureType("intarray");

		sp<RecordArray> spRecordArray=recordArray();
		if(spRecordArray.isNull() || a_index>=spRecordArray->length())
		{
			return 0;
		}

		Record record=spRecordArray->getRecord(a_index);
		std::vector<I32> vertexArray=m_vertexIndexAccessor(record);

		if(a_subIndex>=vertexArray.size())
		{
			return 0;
		}

		return vertexArray[a_subIndex];
	}

	ensureType("integer");

	sp<RecordArray> spRecordArray=recordArray();
	if(spRecordArray.isNull() || a_index>=spRecordArray->length())
	{
		return 0;
	}

	Record record=spRecordArray->getRecord(a_index);
	if(record.isValid())
	{
		return m_accessorInteger(record);
	}

	return 0;
}

void SurfaceAccessorBase::append(U32 a_index,I32 a_integer)
{
	ensureType("intarray");

	sp<RecordArray> spRecordArray=recordArray();
	if(spRecordArray.isNull())
	{
		feLog("SurfaceAccessorBase::append(U32,I32) null record array\n");
		return;
	}

	Record record=spRecordArray->getRecord(a_index);
	m_vertexIndexAccessor(record).push_back(a_integer);
}

I32 SurfaceAccessorBase::append(void)
{
	const String layoutName=SurfaceAccessibleBase::elementLayout(m_element);

	sp<Scope> spScope=m_record.layout()->scope();

	spScope->declare(layoutName);

	sp<Layout> spLayout=spScope->lookupLayout(layoutName);

	if(spLayout.isNull())
	{
		feLog("SurfaceAccessorBase::append"
				" no such layout \"%s\" for \"%s\" type \"%s\"\n",
				layoutName.c_str(),m_attrName.c_str(),m_attrType.c_str());
		FEASSERT(FALSE);

		return 0;
	}

	Record record=spScope->createRecord(spLayout);
	m_spRG->add(record);

	sp<RecordArray> spRecordArray=recordArray();
	if(spRecordArray.isValid())
	{
		if(!m_surfacePointRV.record().isValid())
		{
			m_surfacePointRV.bind(spRecordArray);
		}

		return spRecordArray->length();
	}

	return 0;
}

void SurfaceAccessorBase::set(std::vector< std::vector<I32> >& a_rPrimVerts)
{
	if(m_element!=SurfaceAccessibleI::e_primitive ||
			m_attribute!=SurfaceAccessibleI::e_vertices)
	{
		return;
	}

	I32 pointCount=0;

	const U32 primitiveCount=a_rPrimVerts.size();
	for(U32 primitiveIndex=0;primitiveIndex<primitiveCount;primitiveIndex++)
	{
		append();
		FEASSERT(primitiveIndex==count()-1);

		const std::vector<I32>& rVertList=a_rPrimVerts[primitiveIndex];

		const U32 subCount=rVertList.size();
		for(U32 subIndex=0;subIndex<subCount;subIndex++)
		{
			const I32 pointIndex=rVertList[subIndex];

			if(pointCount<pointIndex+1)
			{
				pointCount=pointIndex+1;
			}

			append(primitiveIndex,pointIndex);
		}
	}

	sp<SurfaceAccessorI> spPointPosition=m_spSurfaceAccessibleI->accessor(
			SurfaceAccessibleI::e_point,SurfaceAccessibleI::e_position);

	for(U32 pointIndex=0;pointIndex<pointCount;pointIndex++)
	{
		spPointPosition->append();
	}
}

void SurfaceAccessorBase::set(U32 a_index,U32 a_subIndex,Real a_real)
{
	ensureType("real");

	sp<RecordArray> spRecordArray=recordArray();
	if(spRecordArray.isNull() || a_index>=spRecordArray->length())
	{
		return;
	}

	Record record=spRecordArray->getRecord(a_index);
	if(record.isValid())
	{
		m_accessorReal(record)=a_real;
	}
}

Real SurfaceAccessorBase::real(U32 a_index,U32 a_subIndex)
{
	ensureType("real");

	sp<RecordArray> spRecordArray=recordArray();
	if(spRecordArray.isNull() || a_index>=spRecordArray->length())
	{
		return Real(0);
	}

	Record record=spRecordArray->getRecord(a_index);
	if(record.isValid())
	{
		return m_accessorReal(record);
	}

	return Real(0);
}

void SurfaceAccessorBase::set(U32 a_index,U32 a_subIndex,
	const SpatialVector& a_vector)
{
	ensureType("vector3");

	m_surfacePointRV.setIndex(a_index);

	switch(m_attribute)
	{
		case SurfaceAccessibleI::e_generic:
			break;
		case SurfaceAccessibleI::e_position:
			m_surfacePointRV.at()=a_vector;
			return;
		case SurfaceAccessibleI::e_normal:
			m_surfacePointRV.up()=a_vector;
			return;
		default:
			return;
	}

	if(m_attrName=="uv")
	{
		m_surfacePointRV.uvw()=a_vector;
		return;
	}

	sp<RecordArray> spRecordArray=recordArray();
	if(spRecordArray.isNull() || a_index>=spRecordArray->length())
	{
		return;
	}

	Record record=spRecordArray->getRecord(a_index);
	if(record.isValid())
	{
		m_accessorSpatialVector(record)=a_vector;
	}
}

SpatialVector SurfaceAccessorBase::spatialVector(U32 a_index,U32 a_subIndex)
{
	ensureType("vector3");

	m_surfacePointRV.setIndex(a_index);

	switch(m_attribute)
	{
		case SurfaceAccessibleI::e_generic:
			break;
		case SurfaceAccessibleI::e_position:
			return m_surfacePointRV.at();
		case SurfaceAccessibleI::e_normal:
			return m_surfacePointRV.up();
		default:
			return SpatialVector(Real(0),Real(0),Real(0));
	}

	if(m_attrName=="uv")
	{
		return m_surfacePointRV.uvw();
	}

	sp<RecordArray> spRecordArray=recordArray();
	if(spRecordArray.isNull() || a_index>=spRecordArray->length())
	{
		return SpatialVector(Real(0),Real(0),Real(0));
	}

	Record record=spRecordArray->getRecord(a_index);
	if(record.isValid())
	{
		return m_accessorSpatialVector(record);
	}

	return SpatialVector(Real(0),Real(0),Real(0));
}

BWORD SurfaceAccessorBase::bindInternal(SurfaceAccessibleI::Element a_element,
		const String& a_name)
{
	m_element=a_element;
	m_attrName=a_name;

	if(!m_record.isValid())
	{
		feLog("SurfaceAccessorBase::bindInternal invalid Record\n");
		return FALSE;
	}

	SurfaceModel surfaceModelRV;
	surfaceModelRV.bind(m_record);

	switch(m_element)
	{
		case SurfaceAccessibleI::e_point:
			m_spRG=surfaceModelRV.surfacePointRG();
			break;
		case SurfaceAccessibleI::e_vertex:
			m_spRG=surfaceModelRV.surfaceVertexRG();
			break;
		case SurfaceAccessibleI::e_primitive:
			m_spRG=surfaceModelRV.surfacePrimitiveRG();
			break;
		case SurfaceAccessibleI::e_detail:
			m_spRG=surfaceModelRV.surfaceDetailRG();
			break;
		default:
			feLog("SurfaceAccessorBase::bindInternal invalid element\n");
			return FALSE;
	}

	sp<RecordArray> spRecordArray=recordArray();
	if(spRecordArray.isValid())
	{
		m_surfacePointRV.bind(spRecordArray);
	}


	if(m_attribute==SurfaceAccessibleI::e_vertices)
	{
		sp<Scope> spScope=m_record.layout()->scope();
		m_vertexIndexAccessor.setup(spScope,"spc:verts","intarray");
	}

	return TRUE;
}

void SurfaceAccessorBase::fragmentWith(SurfaceAccessibleI::Element a_element,
	String a_attributeName,String a_groupName)
{
//	feLog("SurfaceAccessorBase::fragmentWith \"%s\" group \"%s\"\n",
//			a_attributeName.c_str(),a_groupName.c_str());

	m_fragmentMap.clear();
	if(m_spSurfaceAccessibleI.isNull())
	{
		feLog("SurfaceAccessorBase::fragmentWith"
				" m_spSurfaceAccessibleI not set\n");
		return;
	}

	sp<SurfaceAccessorI> spFragmentAccessor=m_spSurfaceAccessibleI->accessor(
			a_element,a_attributeName);
	if(spFragmentAccessor.isNull())
	{
		feLog("SurfaceAccessorBase::fragmentWith"
				" access to attribute \"%s\" failed\n",
				a_attributeName.c_str());
		return;
	}

	sp<SurfaceAccessorI> spGroupAccessor;
	if(!a_groupName.empty())
	{
		const SurfaceAccessibleI::Element groupElement=
				(a_element==SurfaceAccessibleI::e_primitive)?
				SurfaceAccessibleI::e_primitiveGroup:
				SurfaceAccessibleI::e_pointGroup;
		spGroupAccessor=
				m_spSurfaceAccessibleI->accessor(groupElement,a_groupName);
	}

	U32 elementCount=spGroupAccessor.isValid()?
			spGroupAccessor->count(): spFragmentAccessor->count();
	for(U32 elementIndex=0;elementIndex<elementCount;elementIndex++)
	{
		const U32 fragmentIndex=spGroupAccessor.isValid()?
				spGroupAccessor->integer(elementIndex): elementIndex;
		String fragment=spFragmentAccessor->string(fragmentIndex);
		Fragment& rFragment=m_fragmentMap[fragment];
		rFragment.m_indexArray.push_back(fragmentIndex);

//		feLog("SurfaceAccessorBase::fragmentWith %d/%d as %d fragment \"%s\"\n",
//				elementIndex,elementCount,fragmentIndex,fragment.c_str());
	}
}

} // namespace
