/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#ifndef __core_Type_h__
#define __core_Type_h__

namespace fe
{

/**	@brief C++ type_info wrapper

	@ingroup core
	*/
class FE_DL_EXPORT TypeInfo
{
	public:
		class		not_initialized {};

					TypeInfo(void);
					TypeInfo(const std::type_info &other);
					TypeInfo(const TypeInfo &other);
		const		std::type_info &ref(void) const;

		TypeInfo	&operator=(const TypeInfo &other);
		bool		operator==(const TypeInfo &other) const;
		bool		operator!=(const TypeInfo &other) const;
		bool		operator<(const TypeInfo &other) const;
		bool		operator>(const TypeInfo &other) const;
		bool		operator>=(const TypeInfo &other) const;
		bool		operator<=(const TypeInfo &other) const;

		bool		isValid(void) const;
	private:
		const std::type_info	*m_pTypeInfo;
};

class FE_DL_EXPORT BaseTypeVector : public Counted
{
	public:
		BaseTypeVector(void)
		{
			m_base = NULL;
			m_size = 0;
		}
virtual	~BaseTypeVector(void)
		{
		}

//virtual	void	*vectorRoot(void) = 0;
//virtual	size_t	itemSize(void) = 0;
virtual	void	resize(unsigned int a_size) {}

		// raw_at cannot be a virtual due to speed issues
		void	*raw_at(unsigned int a_index)
		{
			return (void *)((char *)m_base + m_size*a_index);
		}

#if FE_COUNTED_STORE_TRACKER
const	String	verboseName(void) const
				{	return "BaseTypeVector "+name(); }
#endif

	protected:
		void			*m_base;
		unsigned int	m_size;
};

/**	@brief A class to associate functionality with run time types.

	@ingroup core

	The primary purpose of this class is to associate functionaility such
	as serialization, construction, and destruction with library/application
	managed instances of C++ types.

	This is done by creating a BaseType, typically by using the Type template
	class through TypeMaster.  The necessary functionality is associated by
	creating a BaseType::Info object and attaching it to the BaseType.

	@code
	class Info32 : public BaseType::Info
	{
	virtual	void	output(std::ostream &ostrm, void *instance, t_serialMode mode)
			{
				U32 u32;
				u32 = htonl(*(U32 *)instance);
				ostrm.write((char *)&u32, sizeof(U32));
			}
	virtual	void	input(std::istream &istrm, void *instance, t_serialMode mode)
			{
				U32 u32;
				istrm.read((char *)&u32, sizeof(U32));
				*(U32 *)instance = ntohl(u32);
			}
	virtual	IWORD	iosize(void) { return sizeof(U32); }
	};

	sp<BaseType> spT;
	spT = typeMaster->assertType<U32>("unsigned_integer");
	spT->setInfo(new Info32());

	@endcode

	*/
class FE_DL_EXPORT BaseType : public Counted, public ClassSafe<GlobalHolder>
{
	public:
		class FE_CORE_PORT Info : public Counted
		{
			public:
						typedef enum
						{
							e_binary,
							e_ascii
						} t_serialMode;
						Info(void)
						{
#if FE_COUNTED_STORE_TRACKER
							setName("BaseType::Info");
#endif
#if FE_CODEGEN<=FE_DEBUG && FE_OS==FE_LINUX && FE_HEAP_CHECKABLE
							U8 current_stack;
							if((void*)this>(void*)&current_stack)
								feX("fe::BaseType::Info::Info",
										"Info object not on stack");
#endif
							suppressReport();
						}
		virtual			~Info(void)											{}
		virtual	bool	getConstruct(void)				{ return false; }
		virtual	void	construct(void *instance)							{}
		virtual	void	destruct(void *instance)							{}
		virtual	BWORD	copy(void *instance,const void* source)
						{
							if(iosize()==c_implicit)
								return FALSE;
							memcpy(instance,source,iosize());
							return TRUE;
						}
		virtual	IWORD	output(	std::ostream &ostrm, void *instance,
								t_serialMode mode)							= 0;
		virtual	void	input(	std::istream &istrm, void *instance,
								t_serialMode mode)							= 0;
		virtual	String	print(void *instance)			{ return String(); }
		virtual	IWORD	iosize(void)					{ return c_implicit; }
		virtual	FE_UWORD	alignment(void)					{ return 0; }

		/** implicit variable size. avoid. mainly for internal use. */
		static FE_DL_PUBLIC	const	IWORD	c_implicit;
		/** explicit variable size. network byte order 32 bit unsigned
			integer should follow */
		static FE_DL_PUBLIC	const	IWORD	c_explicit;
		static FE_DL_PUBLIC	const	IWORD	c_noascii;
		static FE_DL_PUBLIC	const	IWORD	c_ascii;

#if FE_COUNTED_STORE_TRACKER
	const	String		verboseName(void) const
						{	return "BaseType::Info "+name(); }
#endif
		};
	public:
						BaseType(void);
virtual					~BaseType(void);
virtual	FE_UWORD		size(void) const;
virtual	void			assign(void *lvalue, void *rvalue)					=0;
virtual	bool			equiv(void *a_a, void *a_b)							=0;
virtual	void			setInfo(const sp<Info> &spInfo);
virtual	void			setInfo(Info *pInfo);
virtual	sp<Info>		getInfo(void);

virtual	bool			getConstruct(void) const;
virtual	void			construct(void *instance);
virtual	void			destruct(void *instance);
virtual	BWORD			copy(void *instance,const void* source);
virtual	TypeInfo		&typeinfo(void);
virtual	sp<BaseTypeVector>	createVector(void)								=0;


#if FE_COUNTED_STORE_TRACKER
const	String			verboseName(void) const		{ return "Type "+name(); }
#endif

	protected:
		FE_UWORD		m_size;
		sp<Info>		m_spInfo;
		TypeInfo		m_typeInfo;
};


template <class T>
class FE_DL_EXPORT TypeInfoInitialized : public BaseType::Info
{
	public:
				TypeInfoInitialized(void)
				{
#if FE_COUNTED_STORE_TRACKER
					setName(typeid(T).name());
#endif
				}
virtual			~TypeInfoInitialized(void)		{ }
virtual	bool	getConstruct(void)				{ return true; }
virtual	void	construct(void *instance)		{ *((T *)instance)=T(0); }
virtual	IWORD	output(	std::ostream &ostrm, void *instance,
						t_serialMode mode)
				{
					if(mode == e_binary) { return 0; }
					else { return c_noascii; }
				}
virtual	void	input(	std::istream &istrm, void *instance,
						t_serialMode mode)		{ }
virtual	IWORD	iosize(void)					{ return sizeof(T); }
};

template <class T>
class FE_DL_EXPORT TypeInfoConstructable : public BaseType::Info
{
	public:
				TypeInfoConstructable(void)
				{
#if FE_COUNTED_STORE_TRACKER
					setName(typeid(T).name());
#endif
				}
virtual			~TypeInfoConstructable(void)	{ }
virtual	bool	getConstruct(void)				{ return true; }
virtual	void	construct(void *instance)		{ new(instance)T; }
virtual	void	destruct(void *instance)		{ ((T *)instance)->~T(); }
virtual	IWORD	output(	std::ostream &ostrm, void *instance,
						t_serialMode mode)
				{
					if(mode == e_binary) { return 0; }
					else { return c_noascii; }
				}
virtual	void	input(	std::istream &istrm, void *instance,
						t_serialMode mode) { }
};

template <class T>
class FE_DL_EXPORT TypeInfoMemcpyIO : public BaseType::Info
{
	public:
virtual	IWORD	output(std::ostream &ostrm, void *instance, t_serialMode mode);
virtual	void	input(std::istream &istrm, void *instance, t_serialMode mode);
virtual	IWORD	iosize(void);
};

#if 0
template <class T>
class FE_DL_EXPORT TypeVector : public BaseTypeVector
{
	public:
		TypeVector(void)
		{
			m_base = NULL;
			m_vector = NULL;
			m_size = sizeof(T);
		}
virtual	~TypeVector(void)
		{
			if(m_vector)
			{
				deallocate(m_vector);
			}
		}

virtual void    resize(unsigned int a_size)
		{
			if(m_vector && a_size == 0)
			{
				deallocate(m_vector);
				m_vector = NULL;
				m_base = NULL;
				return;
			}
			if(!m_vector)
			{
				m_vector = (T *)allocate(a_size*m_size);
			}
			else
			{
				m_vector = (T *)reallocate((void *)m_vector, a_size*m_size);
			}
			m_base = m_vector;
		}

virtual void    *raw_at(unsigned int a_index)
		{
			return (void *)&m_vector[a_index];
		}

		T &at(unsigned int a_index)
		{
			return m_vector[a_index];
		}

	private:
		T				*m_vector;
};
#endif


#if 1
template <class T>
class FE_DL_EXPORT TypeVector : public BaseTypeVector
{
	public:
		TypeVector(void)
		{
			m_base = NULL;
			m_size = sizeof(T);
			//TODO: 1000 seems grossly arbitrary
			m_vector.reserve(1000);
		}
virtual	~TypeVector(void)
		{
#if FE_COUNTED_STORE_TRACKER
			if(m_base)
			{
				deregisterRegion(m_base);
			}
#endif
		}

virtual void    resize(unsigned int a_size)
		{
#if FE_COUNTED_STORE_TRACKER
			if(m_base)
			{
				deregisterRegion(m_base);
			}
#endif

			m_vector.resize(a_size);
			if(a_size > 0)
			{
				m_base = (void *)&(m_vector[0]);
			}

#if FE_COUNTED_STORE_TRACKER
			registerRegion(m_base,a_size*m_size);
#endif
		}

virtual void    *raw_at(unsigned int a_index)
		{
			return (void *)&m_vector[a_index];
		}

		T &at(unsigned int a_index)
		{
			return m_vector[a_index];
		}

#if FE_COUNTED_STORE_TRACKER
const	String	verboseName(void) const		{ return "TypeVector "+name(); }
#endif

	private:
		std::vector<T>	m_vector;
};


// work around stupid STL decision to specialize std::vector<bool>
template <>
class FE_DL_EXPORT TypeVector<bool> : public BaseTypeVector
{
	public:
		TypeVector(void)
		{
			m_base = NULL;
			m_vector = NULL;
			m_size = sizeof(bool);
		}
virtual	~TypeVector(void)
		{
			if(m_vector)
			{
				deallocate(m_vector);
			}
		}

virtual void    resize(unsigned int a_size)
		{
			if(m_vector && a_size == 0)
			{
				deallocate(m_vector);
				m_vector = NULL;
				m_base = NULL;
				return;
			}
			if(!m_vector)
			{
				m_vector = (bool *)allocate(a_size*m_size);
			}
			else
			{
				m_vector = (bool *)reallocate((void *)m_vector, a_size*m_size);
			}
			m_base = m_vector;
		}

virtual void    *raw_at(unsigned int a_index)
		{
			return (void *)&m_vector[a_index];
		}

		bool &at(unsigned int a_index)
		{
			return m_vector[a_index];
		}

	private:
		bool				*m_vector;
};
#endif

/**	@brief A class to associate functionality with run time types.

	@ingroup core

	See BaseType for details.
	*/
template <class T>
class FE_DL_EXPORT Type : public BaseType
{
	public:
						Type(void);
virtual					~Type(void);
virtual		void		assign(void *lvalue, void *rvalue);
virtual		bool		equiv(void *a_a, void *a_b);
static		Type<T>*	create(void);
virtual		sp<BaseTypeVector>	createVector(void);
};

template <class T>
inline sp<BaseTypeVector> Type<T>::createVector(void)
{
	sp< BaseTypeVector > spBTV(new TypeVector<T>);
	return spBTV;
}

template <>
inline sp<BaseTypeVector> Type<void>::createVector(void)
{
	sp< BaseTypeVector > spBTV(new BaseTypeVector());
	return spBTV;
}


template <class T>
inline Type<T>::Type(void)
{
	m_size=sizeof(T);
	m_typeInfo = TypeInfo(typeid(T));
#if FE_COUNTED_STORE_TRACKER
	setName(m_typeInfo.ref().name());
#endif
}

template <class T>
inline Type<T>::~Type(void)
{
}

template <class T>
inline void Type<T>::assign(void *lvalue, void *rvalue)
{	*(reinterpret_cast<T *>(lvalue)) = *(reinterpret_cast<T *>(rvalue)); }

template <class T>
inline bool Type<T>::equiv(void *a_a, void *a_b)
{
	return (*(reinterpret_cast<T *>(a_a)) == *(reinterpret_cast<T *>(a_b)));
}

template <class T>
inline Type<T>* Type<T>::create(void)
{
	Type<T>* pType=new Type<T>;
	pType->registerRegion(pType,sizeof(Type<T>));
//	feLog("Type<>::create %p %p %p %s\n",
//			&create,pType,pType->typeinfo().ref().name(),
//			pType->typeinfo().ref().name());
	return pType;
}

template<>
inline void Type<void>::assign(void *lvalue, void *rvalue)
{ }

template<>
inline bool Type<void>::equiv(void *a_a, void *a_b)
{
	return true;
}

template<>
inline Type<void>::Type(void)
{	m_size=0; }

inline FE_UWORD BaseType::size(void) const
{
	SAFEGUARD;
	return m_size;
}

template <class T>
inline IWORD TypeInfoMemcpyIO<T>::output(std::ostream &ostrm, void *instance,
	t_serialMode mode)
{
	if(mode == e_binary)
	{
		ostrm.write((char *)instance, sizeof(T));
		return sizeof(T);
	}
	return c_noascii;
}

template <class T>
inline void TypeInfoMemcpyIO<T>::input(std::istream &ostrm, void *instance,
	t_serialMode mode)
{
	ostrm.read((char *)instance, sizeof(T));
}

template <class T>
inline IWORD TypeInfoMemcpyIO<T>::iosize(void)
{
	return sizeof(T);
}

inline TypeInfo::TypeInfo(void)
{
	m_pTypeInfo = &typeid(not_initialized);
	FEASSERT(m_pTypeInfo);
}

inline TypeInfo::TypeInfo(const std::type_info &other)
{
	m_pTypeInfo = &other;
	FEASSERT(m_pTypeInfo);
}

inline TypeInfo::TypeInfo(const TypeInfo &other)
{
	m_pTypeInfo = other.m_pTypeInfo;
	FEASSERT(m_pTypeInfo);
}

inline TypeInfo &TypeInfo::operator=(const TypeInfo &other)
{
	if(this != &other)
	{
		m_pTypeInfo = other.m_pTypeInfo;
	}
	FEASSERT(m_pTypeInfo);
	return *this;
}

inline const std::type_info &TypeInfo::ref(void) const
{
	FEASSERT(m_pTypeInfo);
	return *m_pTypeInfo;
}

inline bool TypeInfo::isValid(void) const
{
	return (bool)(ref() != typeid(not_initialized));
}

#ifdef GCC_HASCLASSVISIBILITY
inline bool TypeInfo::operator==(const TypeInfo &other) const
{
	const char* n1=ref().name();
	const char* n2=other.ref().name();
	return n1==n2 || !strcmp(n1,n2);
}

inline bool TypeInfo::operator!=(const TypeInfo &other) const
{
	return !(*this==other);
}

inline bool TypeInfo::operator<(const TypeInfo &other) const
{
	const char* n1=ref().name();
	const char* n2=other.ref().name();
	return strcmp(n1,n2)<0;
}

inline bool TypeInfo::operator>(const TypeInfo &other) const
{
	const char* n1=ref().name();
	const char* n2=other.ref().name();
	return strcmp(n1,n2)>0;
}

inline bool TypeInfo::operator>=(const TypeInfo &other) const
{
	return !(*this<other);
}

inline bool TypeInfo::operator<=(const TypeInfo &other) const
{
	return !(*this>other);
}
#else
inline bool TypeInfo::operator==(const TypeInfo &other) const
{
	return ref() == other.ref();
}

inline bool TypeInfo::operator!=(const TypeInfo &other) const
{
	return ref() != other.ref();
}

inline bool TypeInfo::operator<(const TypeInfo &other) const
{
	return ref().before(other.ref());
}

inline bool TypeInfo::operator>(const TypeInfo &other) const
{
	return other.ref().before(ref());
}

inline bool TypeInfo::operator>=(const TypeInfo &other) const
{
	return !(ref().before(other.ref()));
}

inline bool TypeInfo::operator<=(const TypeInfo &other) const
{
	return !(other.ref().before(ref()));
}
#endif

} /* namespace */

#endif /* __core_Type_h__ */
